/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Save File to.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


var EXPORTED_SYMBOLS =
[
    "ArrayTransformer",
    "ChainedTransformer",
    "CommandLabelTransformer",
    "DataSourceTransformer",
    "EmptyArrayTransformer",
    "EntryListTrimmer",
    "FolderEntryTransformer",
    "MaxLengthTransformer",
    "NamedFolderTransformer",
    "RecentFolderTransformer",
    "SwitchTransformer",
];

const Cu = Components.utils;

Cu.import("resource://savefileto/common.jsm");
Cu.import("resource://savefileto/Entry.jsm");


/**
 * Transformer that transforms input value using multiple transfomers in sequence.
 */
function ChainedTransformer(transformer /* ... */)
{
    this.transformers = Array.prototype.slice.call(arguments);
}

ChainedTransformer.prototype =
{
    transform: function(value)
    {
        for each (var transformer in this.transformers)
            value = transformer.transform(value);
        return value;
    }
}


/**
 * Transformer that applies |tranformer| to each element of input array.
 * Transforming results (scalar or array) are flattened into a single result array.
 * Errors from transformers can optionally be ignored.
 */
function ArrayTransformer(transformer, ignoreErrors)
{
    this.transformer = transformer;
    this.ignoreErrors = ignoreErrors;
}

ArrayTransformer.prototype =
{
    transform: function(arr)
    {
        var result = [];
        
        for each (var entry in arr)
            try
            {
                var value = this.transformer.transform(entry);
                if (Array.isArray(value))
                    Array.prototype.push.apply(result, value);
                else
                    result.push(value);
            }
            catch (e)
            {
                if (! this.ignoreErrors)
                    throw e;
                
                Cu.reportError(e);
            }
        
        return result;
    },
}


/**
 * Transformer that returns |replacement| if input is empty array.
 */
function EmptyArrayTransformer(replacement)
{
    this.replacement = replacement;
}

EmptyArrayTransformer.prototype =
{
    transform: function(arr)
    {
        return ((arr.length > 0) ? arr : this.replacement);
    }
}


/**
 * Transforms array so that its length is |maxLength| or smaller.
 * If input array is larger, extra elements are trimmed away from the end.
 *
 * @param maxLength.value {int} maximum length of the returned array.
 */
function MaxLengthTransformer(maxLength)
{
    this.maxLength = maxLength;
}

MaxLengthTransformer.prototype =
{
    transform: function(arr)
    {
        var maxLength = this.maxLength.value;
        var result = ((arr.length <= maxLength) ? arr : arr.slice(0, maxLength));
        return result;
    }
}


/**
 * Transformer that transforms input using another transformer based on one of its property.
 */
function SwitchTransformer(prop, transformers)
{
    this.prop = prop;
    this.transformers = transformers;
}

SwitchTransformer.prototype =
{
    transform: function(entry)
    {
        if (! (entry && (this.prop in entry)))
            throw new Error("entry has no property: '" + this.prop + "' " + entry);
        
        var type = entry[this.prop];
        if (! (type in this.transformers))
            throw new Error("no transformer found for type '" + type + "'");
        
        var transformer = this.transformers[type];
        return (transformer ? transformer.transform(entry) : entry);
    }
}


/**
 * "Transformer" that returns data source's entries.
 */
function DataSourceTransformer(dataSource)
{
    this.dataSource = dataSource;
}

DataSourceTransformer.prototype =
{
    transform: function()
    {
        return this.dataSource.entries;
    }
}


/**
 * Transforms raw folder entry to FolderEntry.
 */
function FolderEntryTransformer(compare, filter)
{
    this.compare = compare;
    this.filter = filter;
}

FolderEntryTransformer.prototype =
{
    transform: function(entry)
    {
        return new FolderEntry(entry, this.compare, this.filter);
    }
}


/**
 * Transformer FolderEntry to raw folder entry.
 */
function RecentFolderTransformer()
{
    // empty
}

RecentFolderTransformer.prototype =
{
    transform: function(folder)
    {
        return {type: "folder", label: folder.pathLabel, path: folder.path, menu: false};
    }
}


/**
 * Transformer that creates raw folder entries for named folders.
 *
 * {value: <folder id>, menu: <menu>} ->
 * {type: "folder", menu: <menu>, path: <folder's path>, label: <label from bundle using folder id as the key>}
 * 
 * Note that transform() will throw,
 * if folder isn't valid or the bundle doesn't contain the key and throws.
 *
 * @param bundle {nsIStringBundle}
 */
function NamedFolderTransformer(bundle)
{
    this.bundle = bundle;
}

NamedFolderTransformer.prototype =
{
    transform: function(entry)
    {
        var value = entry.value;
        if (! value)
            throw new Error("entry has no value: " + entry);
        
        var file = getNamedDirectory(value);
        if (! file)
            throw new Error("invalid folder name: " + value);
        
        entry =
        {
            type: "folder",
            label: this.bundle.GetStringFromName(value),
            path: file.path,
            menu: (!! entry.menu)
        }
        
        return entry;
    }
}


/**
 * Transformer that adds label for a command entry.
 *
 * {type: <type>, command: <command>} ->
 * {type: <type>, command: <command>, label: <get value from bundle using command as the key>}
 * 
 * Note that transform() will throw,
 * if the bundle doesn't contain the key and throws.
 *
 * @param bundle {nsIStringBundle}
 */
function CommandLabelTransformer(bundle)
{
    this.bundle = bundle;
}

CommandLabelTransformer.prototype =
{
    transform: function(entry)
    {
        if (! ("label" in entry))
        {
            entry =
            {
                type: entry.type,
                label: this.bundle.GetStringFromName(entry.command),
                command: entry.command
            }
        }
        
        return entry;
    }
}


/**
 * Transformer that trims hidden entries and consecutive separators from the input.
 */
function EntryListTrimmer()
{
    // empty
}

EntryListTrimmer.prototype =
{
    transform: function(entries)
    {
        var result = [];
        var wasPrevSeparator = true;
        
        // strip hidden entries and consecutive separators
        for each (var entry in entries)
        {
            if (entry.hidden)
                continue;
            
            var isEntrySeparator = (entry.type == "separator");
            if (isEntrySeparator && wasPrevSeparator)
                continue;
            
            result.push(entry);
            wasPrevSeparator = isEntrySeparator;
        }
        
        // remove trailing separator
        if (wasPrevSeparator && (result.length > 0))
            result.pop();
        
        return result;
    }
}
